#!/usr/share/ucs-test/runner /usr/share/ucs-test/selenium-pytest -s -l -vvv
## -*- coding: utf-8 -*-
## desc: Check test app login, get access token and verify self disclosure user metadata
## tags: [ucsschool-id-broker]
## packages: [id-broker-provisioning-api-plugin]
## exposure: safe


import os
import time

import ldap
import pytest
import requests
from selenium.common.exceptions import TimeoutException
from utils import get_self_disclosure_metadata, get_token_from_test_app_logon, get_traeger_ldap_admin_con

from univention.admin.rest.client import UDM
from univention.testing.utils import retry_on_error


@pytest.fixture(scope="module")
def impersonation_users(udm):
    lo_t2 = get_traeger_ldap_admin_con("Traeger2")
    udm_mod = udm.get("users/user")
    impersonation_users = {}
    traegers = {
        "Traeger1": f"{os.environ.get('UCS_ENV_TRAEGER1_FQDN')}",
        "Traeger2": f"{os.environ.get('UCS_ENV_TRAEGER2_FQDN')}",
    }
    traeger1_kelvin = f"https://{traegers['Traeger1']}/ucsschool/kelvin"
    traeger1_token = requests.post(
        f"{traeger1_kelvin}/token",
        data={
            "grant_type": "password",
            "username": "Administrator",
            "password": open("/var/lib/ucs-test/Traeger1.secret").read().strip(),
        },
        verify=False,
    ).json()["access_token"]
    traeger2_kelvin = f"https://{traegers['Traeger2']}/ucsschool/kelvin"
    traeger2_token = requests.post(
        f"{traeger2_kelvin}/token",
        data={
            "grant_type": "password",
            "username": "Administrator",
            "password": open("/var/lib/ucs-test/Traeger2.secret").read().strip(),
        },
        verify=False,
    ).json()["access_token"]

    # Avoid id connector erorrs due to the duplicate entry_uuid by using a non school user
    udm_t2 = UDM.http(
        f"http://{traegers['Traeger2']}/univention/udm/",
        "Administrator",
        open("/var/lib/ucs-test/Traeger2.secret").read().strip(),
    )
    user_module_t2 = udm_t2.get("users/user")
    obj = user_module_t2.new()
    obj.properties["username"] = "impersonating_user"
    obj.properties["lastname"] = "Impersonator"
    obj.properties["firstname"] = "Eve"
    obj.properties["password"] = "univention123456789"
    obj.save()
    impersonation_users["impersonating_user"] = {"dn": obj.dn}

    impersonation_users["impersonated_user_t2"] = requests.post(
        f"{traeger2_kelvin}/v1/users/",
        json={
            "name": "impersonated_user_t2",
            "firstname": "Alice",
            "lastname": "Innocent",
            "school": f"{traeger2_kelvin}/v1/schools/ou1",
            "password": "univention123456789",
            "roles": [f"{traeger2_kelvin}/v1/roles/teacher"],
            "record_uid": "impersonated_user_t2",
        },
        verify=False,
        headers={"Authorization": f"Bearer {traeger2_token}"},
    ).json()
    assert "dn" in impersonation_users["impersonated_user_t2"]
    impersonation_users["impersonated_user_t1"] = requests.post(
        f"{traeger1_kelvin}/v1/users/",
        json={
            "name": "impersonated_user_t1",
            "firstname": "Bob",
            "lastname": "Innocent",
            "school": f"{traeger1_kelvin}/v1/schools/ou1",
            "password": "univention123456789",
            "roles": [f"{traeger1_kelvin}/v1/roles/teacher"],
            "record_uid": "impersonated_user_t1",
        },
        verify=False,
        headers={"Authorization": f"Bearer {traeger1_token}"},
    ).json()
    assert "dn" in impersonation_users["impersonated_user_t1"]
    for user in [
        "Traeger2-impersonating_user",
        "Traeger2-impersonated_user_t2",
        "Traeger1-impersonated_user_t1",
    ]:
        print(f"Waiting for {user}")
        retry_on_error(lambda: list(udm_mod.search(f"uid={user}"))[0])
    time.sleep(6)  # sleep for sddb
    entry_uuid_old = lo_t2.get(impersonation_users["impersonating_user"]["dn"], ("entryUUID",))[
        "entryUUID"
    ][0]
    yield impersonation_users
    entry_uuid_new = lo_t2.get(impersonation_users["impersonating_user"]["dn"], ("entryUUID",))[
        "entryUUID"
    ][0]
    lo_t2.modify(
        impersonation_users["impersonating_user"]["dn"],
        [("entryUUID", entry_uuid_new, entry_uuid_old)],
        serverctrls=[ldap.controls.simple.RelaxRulesControl()],
    )
    traeger1_token = requests.post(
        f"{traeger1_kelvin}/token",
        data={
            "grant_type": "password",
            "username": "Administrator",
            "password": open("/var/lib/ucs-test/Traeger1.secret").read().strip(),
        },
        verify=False,
    ).json()["access_token"]
    traeger2_token = requests.post(
        f"{traeger2_kelvin}/token",
        data={
            "grant_type": "password",
            "username": "Administrator",
            "password": open("/var/lib/ucs-test/Traeger2.secret").read().strip(),
        },
        verify=False,
    ).json()["access_token"]
    udm_t2 = UDM.http(
        f"http://{traegers['Traeger2']}/univention/udm/",
        "Administrator",
        open("/var/lib/ucs-test/Traeger2.secret").read().strip(),
    )
    user_module_t2 = udm_t2.get("users/user")
    obj = user_module_t2.get(impersonation_users["impersonating_user"]["dn"])
    obj.delete()
    requests.delete(
        f"{traeger2_kelvin}/v1/users/impersonated_user_t2",
        verify=False,
        headers={"Authorization": f"Bearer {traeger2_token}"},
    )
    requests.delete(
        f"{traeger1_kelvin}/v1/users/impersonated_user_t1",
        verify=False,
        headers={"Authorization": f"Bearer {traeger1_token}"},
    )


@pytest.mark.parametrize(
    "uid",
    [
        "Traeger1-stud1",
        "Traeger1-teach1",
        "Traeger2-stud1",
        "Traeger2-teach1",
    ],
)
def test_test_app_login(selenium, udm, uid):
    traeger = uid.split("-")[0]
    username = uid.split("-", 1)[1]
    kdc_idp_hint = traeger.lower()
    password = "univention"

    tokens = get_token_from_test_app_logon(selenium, username, password, kdc_idp_hint)

    # Check the time on all systems, if this fails repeatedly!
    user_metadata = get_self_disclosure_metadata(tokens["access_token"])

    udm_mod = udm.get("users/user")
    broker_user = list(udm_mod.search(f"uid={uid}"))[0]
    assert user_metadata["username"] == "{}.{}".format(
        broker_user.props.firstname, broker_user.props.lastname
    )
    assert user_metadata["firstname"] == broker_user.props.firstname
    assert user_metadata["lastname"] == broker_user.props.lastname
    assert user_metadata["school_id"] in [
        school.replace(f"{traeger}-", "") for school in broker_user.props.school
    ]
    assert user_metadata["school_authority"] == broker_user.props.ucsschoolSourceUID
    assert user_metadata["type"] in [role.split(":")[0] for role in broker_user.props.ucsschoolRole]


def test_user_impersonation_single_idp(selenium, udm, impersonation_users):
    # Test that logins from traegerX are trusted unconditional for users from traegerX
    lo_t2 = get_traeger_ldap_admin_con("Traeger2")
    traeger = "Traeger2"
    kdc_idp_hint = traeger.lower()
    password = "univention123456789"

    # Lets use a user from traeger2 to impersonate another user of traeger2.
    # We can do this by manipulating the entryUUID.
    # We fully trust the IdP of traeger2 for users from traeger2, so this should work.
    # This first test is to make sure this works in theory.

    entry_uuid_target = lo_t2.get(impersonation_users["impersonated_user_t2"]["dn"], ("entryUUID",))[
        "entryUUID"
    ][0]
    entry_uuid = lo_t2.get(impersonation_users["impersonating_user"]["dn"], ("entryUUID",))["entryUUID"][
        0
    ]

    lo_t2.modify(
        impersonation_users["impersonating_user"]["dn"],
        [("entryUUID", entry_uuid, entry_uuid_target)],
        serverctrls=[ldap.controls.simple.RelaxRulesControl()],
    )

    tokens = get_token_from_test_app_logon(selenium, "impersonating_user", password, kdc_idp_hint)

    # Check the time on all systems, if this fails repeatedly!
    user_metadata = get_self_disclosure_metadata(tokens["access_token"])

    udm_mod = udm.get("users/user")
    # See, we used impersonating_user to login but the broker returns impersonated_user_t2
    broker_user = list(udm_mod.search(f"uid={traeger}-impersonated_user_t2"))[0]
    assert user_metadata["username"] == "{}.{}".format(
        broker_user.props.firstname, broker_user.props.lastname
    )
    assert user_metadata["firstname"] == broker_user.props.firstname
    assert user_metadata["lastname"] == broker_user.props.lastname
    assert user_metadata["school_id"] in [
        school.replace(f"{traeger}-", "") for school in broker_user.props.school
    ]
    assert user_metadata["school_authority"] == broker_user.props.ucsschoolSourceUID
    assert user_metadata["type"] in [role.split(":")[0] for role in broker_user.props.ucsschoolRole]


def test_user_impersonation_cross_idp(selenium, udm, impersonation_users):
    # Test that logins from traegerX are only trusted for users from traegerX and not for any other
    lo_t1 = get_traeger_ldap_admin_con("Traeger1")
    lo_t2 = get_traeger_ldap_admin_con("Traeger2")
    traeger = "Traeger2"
    kdc_idp_hint = traeger.lower()
    password = "univention123456789"

    entry_uuid = lo_t2.get(impersonation_users["impersonating_user"]["dn"], ("entryUUID",))["entryUUID"][
        0
    ]

    entry_uuid_target = lo_t1.get(impersonation_users["impersonated_user_t1"]["dn"], ("entryUUID",))[
        "entryUUID"
    ][0]
    lo_t2.modify(
        impersonation_users["impersonating_user"]["dn"],
        [("entryUUID", entry_uuid, entry_uuid_target)],
        serverctrls=[ldap.controls.simple.RelaxRulesControl()],
    )
    # This must fail!
    with pytest.raises(TimeoutException):
        get_token_from_test_app_logon(selenium, "impersonating_user", password, kdc_idp_hint)
    assert "Invalid username or password." in selenium.page_source
